/*<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>*/
/*
	Project Name:   alt2orc
	File Name:      alt2orc.cpp

  <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

	SSS, EMI, DTU, 23.06.1997

  <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
	Description:
	Program for updating OrCAD input file with pin-information from Altera
	output file.
  <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>*/
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
/* Global Declarations */
struct InpFields
{
	char IPinName[30];
	int  IPinNo;
	char IPinType[10];
} Inp[300];

struct OutpHeader
{
	char HdrLine[5][80];
}OutpHdr;
struct OutpFields
{
	char PinLabel[10];
	int  OPinNo;
	char OPinType[10];
	char OPinName[30];
	char PinStatus[20];
} Outp[300], OutNew[300];
/* PinStatus: <blank=no change>, Modified, New, Removed */

static int  ReadInputFile(FILE *InFile, FILE *LogFile);
static int  ReadOutputFile(FILE *OutFile, FILE *LogFile);
static void InitOutp(void);
static int  UpdatePinNumbers(FILE *LogFile);
static int  WriteOutputFile(FILE *OutFile, FILE *LogFile);

void main (int argc, char *argv[])
{
	FILE *InFile, *OutFile, *LogFile;

	printf("PINUPD version 1.02\n\n");
	if (argc <= 2)
	{
		printf("Updates OrCAD input files using pin information from Altera <name>.fit file\n\n");
		printf("Activation:\n");
		printf("        PINUPD <Altera output file> <OrCAD input file> [log file]\n\n");
		printf("Expected format of Altera output file:\n");
		printf("A random number of header lines, ENDING WITH THE LINES:\n");
		printf("BEGIN\n\nDEVICE = ...\n\n");
		printf("Followed by pin information lines, NB! NO BLANK LINES ALLOWED:\n");
		printf("<pin name> ':' <pin type> '=' <pin number> ;'\n");
		printf("<pin name>:   up to 25 characters, BLANKS NOT ALLOWED!\n");
		printf("<pin number>: 0 to 299\n");
		printf("<pin type>:   INPUT_PIN | OUTPUT_PIN | BIDIR_PIN | LOCATION \n\n");
		printf("The program stops reading from the Altera file when a line containing\n");
		printf("the pin type LOCATION is read or when syntax is violated.\n");
		printf("Press any key ...");
		(void)getch();
		printf("\nExpected format of OrCAD input file:\n");
		printf("5 LINES of header information - NO MORE, NO LESS!\n");
		printf("Followed by pin information lines, NB! NO BLANK LINES ALLOWED:\n");
		printf("<pin label> <pin number> <OrCad pin type> '<pin name>'\n");
		printf("<pin label>:      Tnnn | Bnnn | Lnnn | Rnnn ,nnn being a number, which,\n");
		printf("                  together with T,B,L, or R forms a unique identifier.\n");
		printf("<pin number>:     0 to 299\n");
		printf("<OrCAD pin type>: affected by PINUPD:   hiZ | IN | OUT \n");
		printf("                  unaffected by PINUPD: PWR | OC | I/O \n");
		printf("<pin name>:       up to 25 characters, BLANKS NOT ALLOWED!\n\n");
		printf("The program stops reading from the file when EOF is reached\n");
		printf("or an error is detected. An empty line is considered an error!\n\n");
		printf("Default log file name is PINUPD.LOG.\n\n");
		printf("Press any key ...");
		(void)getch();
		printf("\nNew lines appended to the OrCAD file will be labelled Lnnn if the pin type\n");
		printf("was INPUT, otherwise they will be labelled Rnnn, nnn continuing from highest\n");
		printf("value used with L or R.\n");
		printf("Lines are deleted from the OrCAD file if the following conditions are met:\n");
		printf("   - Pin name does not exist in the Altera file AND\n");
		printf("   - OrCAD pin type is hiZ, IN, or OUT\n");
		printf("Deleted lines are written to the log file. Can be recovered using an editor.\n");
		printf("The OrCAD input file is rewritten. New elements are appended.\n");
		exit(1);
	}
	if (argc == 3)
		LogFile = fopen("PINUPD.LOG","w");
	else
		LogFile = fopen(argv[3],"w");
	if (LogFile == NULL)
	{
		printf("Cannot open/create log file!\n");
		exit(1);
	}
	InitOutp();
	InFile = fopen(argv[1],"r");
	if (InFile == NULL)
	{
		printf("Cannot open input file %s\n",argv[1]);
		fprintf(LogFile,"Cannot open input file %s\n",argv[1]);
		fclose(LogFile);
		exit(1);
	}
	if (ReadInputFile(InFile, LogFile) != 0)
	{
		fclose(InFile);
		fclose(LogFile);
		exit(1);
	}
	fclose(InFile);
	OutFile = fopen(argv[2],"r");
	if (OutFile == NULL)
	{
		printf("Cannot open output file %s!\n",argv[2]);
		fprintf(LogFile,"Cannot open output file %s!\n",argv[2]);
		fclose(LogFile);
		exit(1);
	}
	if (ReadOutputFile(OutFile, LogFile) != 0)
	{
		fclose(OutFile);
		fclose(LogFile);
		exit(1);
	}
	fclose(OutFile);
	if (UpdatePinNumbers(LogFile) != 0)
	{
		fclose(OutFile);
		fclose(LogFile);
		exit(1);
	}
	OutFile = fopen(argv[2],"w");
	(void)WriteOutputFile(OutFile, LogFile);
	fclose(OutFile);
	fclose(LogFile);
	return;
}

static int  ReadInputFile(FILE *InFile, FILE *LogFile)
{
	/*
	Read past header of file: Words to look for are BEGIN and DEVICE.
	Each valid line contains a ':' and either INPUT_PIN, OUTPUT_PIN, or
	BIDIR_PIN.
	Contents of valid lines are stored sequentially in struct Inp.
	*/
	char InpLine[80],PinNa[80],DD[80],D1[80],PinNo[80],D2[80],PType[80],D3[80],
					DN[80],DT[80];
	int i=0,j=0,p=0, NoOfBegins=0;

	while (NoOfBegins < 1)
	{
		if (fgets(InpLine,80,InFile) == NULL)
		{
			printf("Unexpected end of file!\n");
			fprintf(LogFile,"Unexpected end of file!\n");
			exit(1);
		}
		p = sscanf(InpLine,"%s",D1);
		if (stricmp(D1, "BEGIN") == 0)
			NoOfBegins++;
	}
	for (j=0; j<3; j++)
	  fgets(InpLine,80,InFile);
	for (;;)
	{
		fgets(InpLine,80,InFile);
		p = sscanf(InpLine,"%s %s %s %s %6s %s",PinNa,D1,PType,D2,PinNo,D3);
		if ((p != 6) || (stricmp(PType,"LOCATION") == 0) ||
			 (stricmp(D1,":") != 0) || (stricmp(D2,"=") != 0))
		{
			printf("Finished reading input file. %d lines read.\n",i);
			fprintf(LogFile,"Finished reading input file. %d lines read.\n",i);
			return 0;
		}
		p = sscanf(InpLine,"%s %s %s %s %6s %s",DN,D1,DT,D2,DD,D3);
		if (p != 6)
		{
			printf("Conversion error in line %d:\n%s",i, InpLine);
			fprintf(LogFile,"Conversion error in line %d:\n%s",i, InpLine);
			return 1;
		}
		for (j=1; j<(strlen(DN)-1); j++)
		  Inp[i].IPinName[j-1] = DN[j];
		Inp[i].IPinName[j-1] = '\0';
		for (j=0; j<(strlen(DT)-4); j++)
		  Inp[i].IPinType[j] = DT[j];
		Inp[i].IPinNo = atoi(DD);
		i++;
	}
}
static int  ReadOutputFile(FILE *OutFile, FILE *LogFile)
{
	char InpLine[80],DD[80],X1[80],X2[80],X3[80],XVa[80];
	int i=0, p=0, n=0;

	for (i=0; i<5; i++)
		fgets(OutpHdr.HdrLine[i],80,OutFile);
	i=0;
	for (;;)
	{
		if (fgets(InpLine,80,OutFile) == NULL)
		{
			printf("Finished reading output file. %d lines read.\n",i);
			fprintf(LogFile,"Finished reading output file. %d lines read.\n",i);
			break;
		}
		p = sscanf(InpLine,"%s %s %s %s",X1,X2,X3,XVa);
		if (p!=4)
		{
			printf("Error during processing of output file line:\n%s",InpLine);
			fprintf(LogFile,"Error during processing of output file line:\n%s",InpLine);
			return 1;
		}
		p = sscanf(InpLine,"%s %s %s %s",Outp[i].PinLabel,DD,Outp[i].OPinType,X1);
		if (p != 4)
		{
			printf("Conversion error in line %d:\n%s",i, InpLine);
			fprintf(LogFile,"Conversion error in line %d:\n%s",i, InpLine);
			return 1;
		}
		Outp[i].OPinNo = atoi(DD);
		for (n=1;n<strlen(X1)-1;n++)
			Outp[i].OPinName[n-1] = X1[n];
		Outp[i].OPinName[n-1] = '\0';
		for (n=strlen(Outp[i].PinLabel); n<=5; n++)
			Outp[i].PinLabel[n] = ' ';
		Outp[i].PinLabel[n] = '\0';
		i++;
	}
	return 0;
}
static void InitOutp(void)
{
	int i;
	for (i=0; i<300; i++)
	{
		strcpy(Inp[i].IPinName,"    ");
		Inp[i].IPinNo = 0;
		strcpy(Inp[i].IPinType,"    ");

		strcpy(Outp[i].PinLabel,"    ");
		Outp[i].OPinNo    = 0;
		strcpy(Outp[i].OPinType,"   ");
		strcpy(Outp[i].OPinName,"                              ");
		strcpy(Outp[i].PinStatus,"          ");

		strcpy(OutNew[i].PinLabel,"    ");
		OutNew[i].OPinNo    = 0;
		strcpy(OutNew[i].OPinType,"   ");
		strcpy(OutNew[i].OPinName,"                              ");
		strcpy(OutNew[i].PinStatus,"          ");
		/* PinStatus: <blank=no change>, Modified, New, Removed */
	}
}
static int  UpdatePinNumbers(FILE *LogFile)
{
	int i, LMax,RMax,tmp;
	int NewIdx=0,Oi,Ii,found,n;
	/* Find max L-index and max R-index */
	for (i=0,LMax=0,RMax=0;Outp[i].PinLabel[0] != ' ';i++)
	{
		if (Outp[i].PinLabel[0]=='L')
		{
			tmp=atoi(&Outp[i].PinLabel[1]);
			if (tmp>LMax)
				LMax=tmp;
		}
		if (Outp[i].PinLabel[0]=='R')
		{
			tmp=atoi(&Outp[i].PinLabel[1]);
			if (tmp>RMax)
				RMax=tmp;
		}
	}
	/* Update pin numbers */
	for (Ii=0; Inp[Ii].IPinName[0] != ' '; Ii++)
	{
		found = 0;
		for (Oi=0; Outp[Oi].PinLabel[0] != ' '; Oi++)
		{
			if (stricmp(Inp[Ii].IPinName, Outp[Oi].OPinName) == 0)
			{
				if (Outp[Oi].PinStatus[0] != ' ')
				{
					printf("Duplicate pin name in input file: %s\n",Inp[Ii].IPinName);
					fprintf(LogFile,"Duplicate pin name in input file: %s\n",Inp[Ii].IPinName);
				}
				else
				{
					strcpy(Outp[Oi].PinStatus, "Modified");
					Outp[Oi].OPinNo    = Inp[Ii].IPinNo;
					if (stricmp(Inp[Ii].IPinType, "INPUT") == 0)
						strcpy(Outp[Oi].OPinType, "IN");
					else if (stricmp(Inp[Ii].IPinType, "OUTPUT") == 0)
						strcpy(Outp[Oi].OPinType, "OUT");
					else if (stricmp(Inp[Ii].IPinType, "BIDIR") == 0)
						strcpy(Outp[Oi].OPinType, "hiZ");
				}
				found = 1;
				break;
			}
		} /* end for (Oi... */
		if (!found)
		{
			if (stricmp(Inp[Ii].IPinType, "INPUT") == 0)
			{
				OutNew[NewIdx].PinLabel[0] = 'L';
				itoa(++LMax, &OutNew[NewIdx].PinLabel[1],10);
			}
			else
			{
				OutNew[NewIdx].PinLabel[0] = 'R';
				itoa(++RMax, &OutNew[NewIdx].PinLabel[1],10);
			}
			for (n=strlen(OutNew[NewIdx].PinLabel); n<=5; n++)
				OutNew[NewIdx].PinLabel[n] = ' ';
			Outp[i].PinLabel[n] = '\0';
			OutNew[NewIdx].OPinNo = Inp[Ii].IPinNo;
			if      (stricmp(Inp[Ii].IPinType, "INPUT") == 0)
				strcpy(OutNew[NewIdx].OPinType, "IN");
			else if (stricmp(Inp[Ii].IPinType, "OUTPUT") == 0)
				strcpy(OutNew[NewIdx].OPinType, "OUT");
			else if (stricmp(Inp[Ii].IPinType, "BIDIR") == 0)
				strcpy(OutNew[NewIdx].OPinType, "hiZ");
			strcpy(OutNew[NewIdx].OPinName,Inp[Ii].IPinName);
			strcpy(OutNew[NewIdx].PinStatus, "New");
			NewIdx++;
		}
	} /* end for (Ii... */

	printf("Finished updating pin numbers.\n");
	fprintf(LogFile,"Finished updating pin numbers.\n");
	return 0;
}
static int  WriteOutputFile(FILE *OutFile, FILE *LogFile)
{
	int i=0, DelCnt=0, NewCnt=0;

	for (i=0; i<5; i++)
		fprintf(OutFile,"%s",OutpHdr.HdrLine[i]);

	for (i=0;Outp[i].PinLabel[0] != ' ';i++)
	{
		if ((Outp[i].PinStatus[0] == ' ') &&
			((stricmp(Outp[i].OPinType,"IN")        == 0) ||
			 (stricmp(Outp[i].OPinType,"OUT")       == 0) ||
			 (stricmp(Outp[i].OPinType,"hiZ")       == 0))&&
		  !((stricmp(Outp[i].OPinName,"nSP")       == 0) ||
			 (stricmp(Outp[i].OPinName,"MSEL0")     == 0) ||
			 (stricmp(Outp[i].OPinName,"MSEL1")     == 0) ||
			 (stricmp(Outp[i].OPinName,"nSTATUS")   == 0) ||
			 (stricmp(Outp[i].OPinName,"nCONFIG")   == 0) ||
			 (stricmp(Outp[i].OPinName,"DCLK")      == 0) ||
			 (stricmp(Outp[i].OPinName,"CONF_DONE") == 0) ||
			 (stricmp(Outp[i].OPinName,"nTRST")     == 0) ||
			 (stricmp(Outp[i].OPinName,"DATA0")     == 0) ||
			 (stricmp(Outp[i].OPinName,"SDOUT")     == 0)))
		{
			fprintf(LogFile,"Deleted: %.5s     %3d      %s '%s'\n",
						Outp[i].PinLabel, Outp[i].OPinNo, Outp[i].OPinType, Outp[i].OPinName);
			DelCnt++;
		}
		else
			fprintf(OutFile,"%.5s     %3d      %s '%s'\n",
						Outp[i].PinLabel, Outp[i].OPinNo, Outp[i].OPinType, Outp[i].OPinName);
	}
	for (i=0;OutNew[i].PinLabel[0] != ' ';i++)
	{
		fprintf(OutFile,"%.5s     %3d      %s '%s'\n",
					OutNew[i].PinLabel, OutNew[i].OPinNo, OutNew[i].OPinType, OutNew[i].OPinName);
		fprintf(LogFile,"New    : %.5s     %3d      %s '%s'\n",
					OutNew[i].PinLabel, OutNew[i].OPinNo, OutNew[i].OPinType, OutNew[i].OPinName);
		NewCnt++;
	}
	printf("Finished writing output file. %d lines deleted, %d lines added.\n",DelCnt,NewCnt);
	fprintf(LogFile,"Finished writing output file.%d lines deleted, %d lines added.\n",DelCnt,NewCnt);
	return 0;
}
